#ifndef sphere_particle_H
#define sphere_particle_H
#include <cstdlib>
#include <vector>
#include <map>
#include <string>
#include <sstream>
#include <vector>
#include "R2Vector.h"
#include "ellipse_particle.h"

/* -------- Set the static ellipise parameters -----------*/
double const Ea = .10, Eb = .10;

/* -------- Set the window size for the mirror particles-----------*/
double const L = 1.0;

/* -------- Default program parameters --------------------------- */
int NP = 400;                        // particle number
double Pr = 0.025, Pm = 1;    // Particle radius, mass
double total_time =10.000; 
int random_seed = 20;
double dt=0.005;                      // Time between data collection steps
double GLOBAL_T=0;

/* -------- Internal Globals for sphere-particles --------------- */
const double infinite_time = 10E20;    // Used to mark a non-collsion
map <int,bool> inval_links;            // Marks collisions that need to be re-eval
double next_data_t;                    // The next collision to be processed
struct s_pair {int A, B, C; };         // Used by spheres to mark collisions

/* -------- Data collection variables --------------------------- */
vector<double> C12;                        // Data-collection for g(r)
double bins_xspan;                         // Data-collection for g(r)
double delta_hist, top_hist, bottom_hist;  // Data-collection for g(r)
double HM_radius, HM_radius_epsilon, HM_max, HM_deltar;
int    HM_N;
inline int histogram(double x){ return int( (x-HM_radius) / HM_deltar ); }


/* -------- Function declarations and simple functions -------------- */
class sphere_particle;
void invalidate_collision(const int &ID, vector<sphere_particle> &P);
RVector closest_rvector(RVector & A, RVector &B);
void collect_data(vector<sphere_particle> &P);
void print_stats(const sphere_particle &A);
void print_pos(vector<sphere_particle> &P);
double MAX(double x, double y){ if(x>y) return x; return y;}
double MIN(double x, double y){ if(x<y) return x; return y;}

/* -------- Mirror window initilization -------------------------- */
vector<RVector> windows; int w_size;
void set_windows(int xmax,int ymax){
  double wall[3] = {0,2.0*L,-2.0*L};
  for(int i=0;i<xmax;i++)
    for(int j=0;j<ymax;j++)
	windows.push_back(RVector(wall[i],wall[j]));
  // Remove the first element since it is the orginal, not a mirror
  windows.erase(windows.begin());
  w_size = windows.size(); }
void set_windows2D() {set_windows(3,3);}

/* -------- Random number generators ------------------------------- */
double const   PI = 3.14159265;
double uniform() { return (double(rand())/double(RAND_MAX)); }
double gaussian() {
  double x1,x2, w, y1, y2;
  do {
    x1 = 2.0 * uniform() - 1.0;
    x2 = 2.0 * uniform() - 1.0;
    w  = x1 * x1 + x2 * x2;
  } while ( w >= 1.0 );
  
  w = sqrt( (-2.0 * log( w ) ) / w );
  return x1*w; }

/* -------- Sphere particle with velocity and pos.------------------ */
class sphere_particle {
 public: 
  RVector R,V, Ehit;
  double radius, mass, Ephi;
  int ID;
  int color;

  s_pair hit_ID; double hit_t;
  
  
  sphere_particle(const RVector &r, const RVector &v, const double rad, const double M, const int id){
    R=r; V=v;
    radius=rad;
    mass = M;
    ID = id;
    Ehit = RVector(0,0); Ephi=0;
    color = 0;
    hit_t = infinite_time;
  }
  
  sphere_particle& operator=(const sphere_particle& B) {
    cout << "ASSIGNMENT not coded! killing! " << endl;
    exit(37);
    return *this; }

  // Copy constructor
  sphere_particle(const sphere_particle &B){
    R=B.R; V=B.V;
    radius = B.radius;
    mass   = B.mass;
    ID     = B.ID;
    hit_ID = B.hit_ID;
    hit_t  = B.hit_t;
    Ehit   = B.Ehit;
    Ephi   = B.Ephi;
    color  = B.color;
  }

  ~sphere_particle() {}
  
  void integrate(const double &dt){
    R.x += dt * V.x;
    R.y += dt * V.y;
  }
};


vector <sphere_particle> P;

/* -------- Sphere-particle functions ---------------------------- */

double wall_hit_time2D(const sphere_particle & A, s_pair &S){
  double tX=0, tY=0; int idX, idY;

  if (A.V.x > 0)    { tX = ( L - A.R.x)/A.V.x; idX = -1;}
  else if (A.V.x<0) { tX = (-L - A.R.x)/A.V.x; idX = -2;}
  else {              tX = infinite_time; }
  
  if (A.V.y > 0)    { tY = ( L - A.R.y)/A.V.y; idY = -3;}
  else if (A.V.y<0) { tY = (-L - A.R.y)/A.V.y; idY = -4;}
  else {              tY = infinite_time; }

  S.A = A.ID;

  if ( tX < tY ) {
    S.B = idX;
    return tX; }
  
  // Known that tY < tX, so dropping conditional
  S.B = idY;
  return tY;
}

// Took off inline
inline double particle_particle_hit(const sphere_particle &A, const sphere_particle &B) {

  // Determine if and when the particles hit 
  RVector dR = A.R-B.R;
  RVector dV = A.V-B.V;
  double dvr = dV.dot(dR);
  double g   = A.radius+B.radius;
  double dv2 = dV.dot(dV);
  double dr2 = dR.dot(dR);
  double d   = pow(dvr,2.0) - dv2*(dr2-pow(g,2.0));

  // Particles never hit if this is true
  if( d<0 || dvr >= 0) return infinite_time;

  return  (-dvr - sqrt(d)) / dv2;
}

double Eimpact_time(const double a, const double b, sphere_particle &A){
  //  return infinite_time;
  double r   = A.radius;
  RVector R0 = A.R;
  RVector V0 = A.V;
 
  double phi1, phi2;

  double m1 = V0.y/V0.x;
  double m2 = R0.y - m1*R0.x;
  
  bool isContact = Eroots(a,b,r,m1,m2,phi1,phi2,0.0);
  if(!isContact) {return E_infinite_time;}

  RVector hit1 = Econtact(a,b,r,phi1);
  RVector hit2 = Econtact(a,b,r,phi2);
  
  double t1 = (hit1.x - R0.x)/V0.x;
  double t2 = (hit2.x - R0.x)/V0.x;

  if(t1<0 and  t2<0) {return E_infinite_time;}

  // STUCK INSIDE!!!
  if(t1<0 xor t2<0) {
    A.color=0; 
    if(t2<0) {
    A.Ehit = hit2;
    A.Ephi = phi2;
    return 0;
    }
    else{
    A.Ehit = hit1;
    A.Ephi = phi1;
    return 0;
    }
  }

  // Normal collision
  if(t2<t1) {
    A.Ehit = hit2;
    A.Ephi = phi2;
    return t2;
  }

  A.Ehit = hit1;
  A.Ephi = phi1;
  return t1;

}

double particle_hit_time2D(const sphere_particle &A, const sphere_particle &B,
			   s_pair &particleID){
  particleID.A = A.ID; 
  particleID.B = B.ID;
  particleID.C = -1;
  double t = particle_particle_hit(A,B), tp=0;
  
  sphere_particle copyA = A;
  for(int w=0;w<windows.size();w++) {
    copyA.R += windows.at(w);
    tp = particle_particle_hit(copyA,B);
    if(t>tp) { t=tp; particleID.C = w; }
    copyA.R -= windows.at(w);
  }

  return t; 
}

void calc_Ecollision(sphere_particle & A) {

  double hit_ellipse = Eimpact_time(Ea,Eb,A);
  if(hit_ellipse < A.hit_t ){
    //A.color = 1;
    A.hit_t  = hit_ellipse;
    
    s_pair ellipseID;
    ellipseID.A = A.ID;
    ellipseID.B = -35;
    A.hit_ID    = ellipseID;
  }
  //cout << "calcE " << ' ' << A.ID << ' ' << A.hit_t << endl;
}

void calc_Wcollision(sphere_particle & A) {
  s_pair wallID;
  double hit_wall = wall_hit_time2D(A,wallID);
  if(hit_wall < A.hit_t) {
    A.hit_ID = wallID;
    A.hit_t  = hit_wall;
    //A.color  = 0;
  }
  
}


void calc_collision(sphere_particle & A, vector<sphere_particle> &P) {
  calc_Wcollision(A);
  
  s_pair particleID;
  double hit_particle=infinite_time;
  
  for(int i=0;i<P.size();i++){
    s_pair tmpID; double t;
    if(i!=A.ID) { t = particle_hit_time2D(A, P.at(i), tmpID);}
    else {t = infinite_time;}

    if(t<hit_particle && t>0) {
      hit_particle = t;
      particleID   = tmpID;
    }
  }
  
  // Determine if it hits the wall or another particle first
  if(hit_particle < A.hit_t) {
    A.hit_ID = particleID;
    A.hit_t  = hit_particle;
    //A.color = 10;
  }
}

void handle_pw_collision(sphere_particle &A) {
  // Particle has passed through the boundry conditions
  //A.color = 0;
  if (A.hit_ID.B < 0) {
    A.color=0;
    if      (A.hit_ID.B == -1) { A.R.x = -L;}
    else if (A.hit_ID.B == -2) { A.R.x =  L;}
    else if (A.hit_ID.B == -3) { A.R.y = -L;}
    else if (A.hit_ID.B == -4) { A.R.y =  L;}
  }
}

void handle_pe_collision(sphere_particle &A) {
 // Particle has hit the ellipse
  //  A.color = 1;

  RVector N_hat( Eb*cos(A.Ephi), Ea*sin(A.Ephi) );
  N_hat = N_hat.unit();
  
  double J = ((-2.0) * A.V.dot(N_hat)) / (N_hat.dot(N_hat) * (1/A.mass));
  A.V += (J/A.mass) * N_hat;
}

void handle_pp_collision(sphere_particle &A, sphere_particle &B, const int w){

  if(w>=0) A.R += windows.at(w);

  RVector dR = A.R - B.R;

  if ( abs(dR.abs()-(A.radius+B.radius)) < .01) {

  RVector dV = A.V - B.V;
  double   J = (-2*dV.dot(dR))/(dR.dot(dR) * ((1.0/A.mass) +(1.0/B.mass)));

  // Collisions need to conserve energy?
  double E = A.V.abs()+B.V.abs();

  A.V += (J/A.mass) * dR;
  B.V -= (J/B.mass) * dR;

  double Ep = A.V.abs()+B.V.abs();

  //cout << E-Ep << endl;
  A.V *= E/Ep;
  B.V *= E/Ep;  
  }

  if(w>=0) A.R -= windows.at(w);
}

int COLLECTION_COUNTER = 0;

void integrate(vector<sphere_particle> &P, double total_t) {

  // Determine if data collection is needed
  while( int(GLOBAL_T/dt) > COLLECTION_COUNTER ) {
    COLLECTION_COUNTER++;
    print_pos(P);
    //collect_data(P);
  }
 
  // Run up to a current time interval
  if( next_data_t < total_t ) {
    for(int i=0;i<P.size();i++) P.at(i).integrate(next_data_t);
    GLOBAL_T   += next_data_t;
    total_t    -= next_data_t;
    next_data_t = dt;
  }
  
  // Run remaining time intervals
  while( next_data_t < total_t ) {
    for(int i=0;i<P.size();i++) P.at(i).integrate(dt);
    GLOBAL_T += dt;
    total_t -= dt;

    // Determine if data collection is needed
    while( int(GLOBAL_T/dt) > COLLECTION_COUNTER ) {
      COLLECTION_COUNTER++;
      print_pos(P);
      //collect_data(P);
    }
  }

  // Finish running out the remainder of the collision
  for(int i=0;i<P.size();i++) P.at(i).integrate(total_t);
  GLOBAL_T += total_t;
  next_data_t -= total_t;
}

void run_simulation(vector<sphere_particle> &P) {
  // Find the next collision
  map<double,s_pair> hits;
  for(int i=0;i<P.size();i++) hits[P.at(i).hit_t] = P.at(i).hit_ID;

  double t = hits.begin()->first;
  s_pair S = hits.begin()->second;

  // Integrate to the next collision
  integrate( P, t );
  
  // Handle the collision and update new collision times
  for(int i=0;i<P.size();i++) {
    P.at(i).hit_t -= t;
    inval_links[i]=1; 
  }
  
  if(S.B==-35) {
    handle_pe_collision( P.at(S.A) );
    P.at(S.A).hit_t = infinite_time;
    invalidate_collision( S.A, P );
  }
  else if(S.B < 0) {
    handle_pw_collision( P.at(S.A) );
    P.at(S.A).hit_t = infinite_time;
    invalidate_collision( S.A, P );
    calc_Ecollision(P.at(S.A));
  }

  else {
    handle_pp_collision(P.at(S.A), P.at(S.B), S.C);

    P.at(S.A).hit_t = infinite_time;
    P.at(S.B).hit_t = infinite_time;

    invalidate_collision( S.A, P );
    invalidate_collision( S.B, P );

    calc_Ecollision(P.at(S.A));
    calc_Ecollision(P.at(S.B));
  }


}


void invalidate_collision(const int &ID, vector<sphere_particle> &P) {
  int a,b;
  for(int i=0;i<P.size();i++) {
    a = P.at(i).hit_ID.A;
    b = P.at(i).hit_ID.B;
    

    if(a == ID || b == ID) {
      if(a>=0 && inval_links[a]) {
	calc_collision(P.at(a), P);
	inval_links[a] = 0;
      }
      if(b>=0 && inval_links[b]) {
	calc_collision(P.at(b), P);
	inval_links[b] = 0;
      }
    }
  }

}

double closest_distance(sphere_particle & A, sphere_particle &B) {
  double deltax = A.R.x - B.R.x;
  double deltay = A.R.y - B.R.y;
  double dx = deltax - L * round(deltax/L);
  double dy = deltay - L * round(deltay/L); 
  return sqrt(dx*dx+dy*dy);
}

RVector closest_rvector(RVector & A, RVector &B) {
  RVector R;
  double deltax = A.x - B.x;
  double deltay = A.y - B.y;
  double dx = deltax - L * round(deltax/L);
  double dy = deltay - L * round(deltay/L); 
  R.x = dx; R.y = dy;
  return R;
}

/* -------- Utility functions for spheres (eg. data collection, print) */
void print_stats(const sphere_particle &A) {
  cout << "ID: " << &A.ID << " R:<" << &A.R << "> V:<" << &A.V << ">" 
       << " hit_ID:(" << &A.hit_ID.A << ' ' << &A.hit_ID.B << ' ' << &A.hit_ID.C 
       << ") t=" << &A.hit_t << endl; }


void collect_data(vector<sphere_particle> &P) {
  //  for(int i=0;i<P.size();i++) cout << P[i].R << ';';
  //cout << endl;
}

void set_randominital2D_V(sphere_particle &P, const double &spread) {
    P.V.x = gaussian()*spread; 
    P.V.y = gaussian()*spread;
}

void print_pos(vector<sphere_particle> &P) {
  for(int i=0;i<P.size();i++) {
    cout << P[i].R << ',' << P[i].color << ';';
  } 
  cout << endl;
}

/* -------- Command-line inputs ------------------------------------ */

inline double convertToDouble(const std::string& s) {
   std::istringstream i(s);
   double x;
   if (!(i >> x))
     cout << "Error in input argument! " << endl;
   return x; }
inline int convertToInt(const std::string& s) {
   std::istringstream i(s);
   int x;
   if (!(i >> x))
     cout << "Error in input argument! " << endl;
   return x; }

#endif


